=== modified file '__init__.py'
--- __init__.py	2018-08-17 20:57:24 +0000
+++ __init__.py	2018-09-09 15:56:12 +0000
@@ -14,8 +14,9 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 
-"""Support for the propose command."""
+"""Management of hosted branches."""
 
 from ...commands import plugin_cmds
 
 plugin_cmds.register_lazy("cmd_propose_merge", ["propose"], __name__ + ".cmds")
+plugin_cmds.register_lazy("cmd_publish", [], __name__ + ".cmds")

=== modified file 'cmds.py'
--- cmds.py	2018-09-09 14:31:18 +0000
+++ cmds.py	2018-09-09 15:56:12 +0000
@@ -23,29 +23,64 @@
     controldir,
     errors,
     msgeditor,
+    urlutils,
     )
 from ...i18n import gettext
 from ...commands import Command
 from ...option import (
     ListOption,
+    Option,
     RegistryOption,
     )
 from ...sixish import text_type
+from ...trace import note
 from . import (
     propose as _mod_propose,
     )
 
 
+def branch_name(branch):
+    if branch.name:
+        return branch.name
+    return urlutils.basename(branch.user_url)
+
+
 class cmd_publish(Command):
     __doc__ = """Publish a branch.
 
     Try to create a public copy of a local branch.
     How this is done depends on the submit branch and where it is
     hosted.
+
+    Reasonable defaults are picked for owner name, branch name and project
+    name, but they can also be overridden from the command-line.
     """
 
-    def run(self, submit_branch=None, directory='.'):
-        pass
+    takes_options = [
+            'directory',
+            Option('owner', help='Owner of the new remote branch.'),
+            Option('project', help='Project name for the new remote branch.'),
+            Option('name', help='Name of the new remote branch.'),
+            ]
+    takes_args = ['submit_branch?']
+
+    def run(self, submit_branch=None, owner=None, name=None, project=None,
+            directory='.'):
+        local_branch = _mod_branch.Branch.open_containing(directory)[0]
+        self.add_cleanup(local_branch.lock_write().unlock)
+        if submit_branch is None:
+            submit_branch = local_branch.get_submit_branch()
+            note(gettext('Using submit branch %s') % submit_branch)
+        submit_branch = _mod_branch.Branch.open(submit_branch)
+        if name is None:
+            name = branch_name(local_branch)
+        hoster = _mod_propose.get_hoster(submit_branch)
+        remote_branch, public_url = hoster.publish(
+                local_branch, submit_branch, name=name, project=project,
+                owner=owner)
+        local_branch.set_push_location(remote_branch.user_url)
+        local_branch.set_public_branch(public_url)
+        note(gettext("Pushed to %s") % public_url)
 
 
 class cmd_propose_merge(Command):

=== modified file 'github.py'
--- github.py	2018-09-09 14:31:18 +0000
+++ github.py	2018-09-09 15:56:12 +0000
@@ -19,7 +19,7 @@
 from .propose import (
     Hoster,
     MergeProposal,
-    MergeProposer,
+    MergeProposalBuilder,
     MergeProposalExists,
     )
 
@@ -76,7 +76,7 @@
         raise NotImplementedError(self.publish)
 
     def get_proposer(self, source_branch, target_branch):
-        return GitHubMergeProposer(source_branch, target_branch)
+        return GitHubMergeProposalBuilder(source_branch, target_branch)
 
     @classmethod
     def is_compatible(cls, branch):
@@ -88,7 +88,7 @@
             return True
 
 
-class GitHubMergeProposer(MergeProposer):
+class GitHubMergeProposalBuilder(MergeProposalBuilder):
 
     def __init__(self, source_branch, target_branch):
         self.source_branch = source_branch

=== modified file 'gitlab.py'
--- gitlab.py	2018-09-09 14:31:18 +0000
+++ gitlab.py	2018-09-09 15:56:12 +0000
@@ -22,16 +22,12 @@
     )
 
 from .propose import (
+    Hoster,
     MergeProposal,
-    MergeProposer,
+    MergeProposalBuilder,
     MergeProposalExists,
     )
 
-from ...lazy_import import lazy_import
-lazy_import(globals(), """
-from gitlab import Gitlab
-""")
-
 
 class DifferentGitLabInstances(errors.BzrError):
 
@@ -44,6 +40,7 @@
 
 
 def connect_gitlab(url):
+    from ....gitlab import Gitlab
     # TODO(jelmer): Support authentication
     return Gitlab(url)
 
@@ -73,7 +70,7 @@
         return True
 
 
-class GitlabMergeProposer(MergeProposer):
+class GitlabMergeProposalBuilder(MergeProposalBuilder):
 
     def __init__(self, source_branch, target_branch):
         self.source_branch = source_branch
@@ -85,9 +82,6 @@
         if self.source_host != self.target_host:
             raise DifferentGitLabInstances(self.source_host, self.target_host)
 
-    @classmethod
-    def is_compatible(cls, target_branch, source_branch):
-
     def get_infotext(self):
         """Determine the initial comment for the merge proposal."""
         info = []

=== modified file 'launchpad.py'
--- launchpad.py	2018-09-09 14:31:18 +0000
+++ launchpad.py	2018-09-09 15:56:12 +0000
@@ -17,12 +17,15 @@
 from __future__ import absolute_import
 
 from .propose import (
+    Hoster,
     MergeProposal,
-    MergeProposer,
+    MergeProposalBuilder,
     MergeProposalExists,
     )
 
 from ... import (
+    branch as _mod_branch,
+    controldir,
     errors,
     hooks,
     urlutils,
@@ -34,25 +37,74 @@
     lp_registration,
     )
 """)
+from ...transport import get_transport
 
 
 # TODO(jelmer): Make selection of launchpad staging a configuration option.
 
-
 class Launchpad(Hoster):
     """The Launchpad hosting service."""
 
-
-class LaunchpadMergeProposer(MergeProposer):
+    @classmethod
+    def is_compatible(cls, branch):
+        return lp_api.LaunchpadBranch.plausible_launchpad_url(branch.user_url)
+
+    def publish(self, local_branch, base_branch, name, project=None, owner=None,
+                revision_id=None, overwrite=False):
+        """Publish a branch to the site, derived from base_branch.
+
+        :param base_branch: branch to derive the new branch from
+        :param new_branch: branch to publish
+        :param name: Name of the new branch on the remote host
+        :param project: Optional project name
+        :param owner: Optional owner
+        :return: resulting branch
+        """
+        # TODO(jelmer): Prevent publishing to development focus
+        launchpad = connect_launchpad()
+        base_branch = lp_api.LaunchpadBranch.from_bzr(
+            launchpad, base_branch)
+        if project is None:
+            project = base_branch.lp.project.name
+        if owner is None:
+            owner = launchpad.me.name
+        # TODO(jelmer): Surely there is a better way of creating one of these URLs?
+        to_transport = get_transport("lp:~%s/%s/%s" % (owner, project, name))
+        try:
+            dir_to = controldir.ControlDir.open_from_transport(to_transport)
+        except errors.NotBranchError:
+            # Didn't find anything
+            dir_to = None
+
+        if dir_to is None:
+            br_to = local_branch.create_clone_on_transport(to_transport, revision_id=revision_id)
+        else:
+            try:
+                br_to = dir_to.push_branch(local_branch, revision_id, overwrite=overwrite).target_branch
+            except errors.DivergedBranches:
+                raise errors.BzrCommandError(gettext('These branches have diverged.'
+                                        '  See "brz help diverged-branches"'
+                                        ' for more information.'))
+        return br_to, ("https://code.launchpad.net/~%s/%s/%s" % (owner, project, name))
+
+    def get_proposer(self, source_branch, target_branch):
+        return LaunchpadMergeProposalBuilder(source_branch, target_branch)
+
+
+def connect_launchpad(lp_instance='production'):
+    service = lp_registration.LaunchpadService(lp_instance=lp_instance)
+    return lp_api.login(service)
+
+
+class LaunchpadMergeProposalBuilder(MergeProposalBuilder):
 
     def __init__(self, source_branch, target_branch, message=None,
-                 reviews=None, staging=None, approve=None, fixes=None):
+                 staging=None, approve=None, fixes=None):
         """Constructor.
 
         :param source_branch: The branch to propose for merging.
         :param target_branch: The branch to merge into.
         :param message: The commit message to use.  (May be None.)
-        :param reviews: A list of tuples of reviewer, review type.
         :param staging: If True, propose the merge against staging instead of
             production.
         :param approve: If True, mark the new proposal as approved immediately.
@@ -64,8 +116,7 @@
             lp_instance = 'staging'
         else:
             lp_instance = 'production'
-        service = lp_registration.LaunchpadService(lp_instance=lp_instance)
-        self.launchpad = lp_api.login(service)
+        self.launchpad = connect_launchpad(lp_instance=lp_instance)
         self.source_branch = lp_api.LaunchpadBranch.from_bzr(
             self.launchpad, source_branch)
         if target_branch is None:
@@ -78,12 +129,6 @@
         self.approve = approve
         self.fixes = fixes
 
-    @classmethod
-    def is_compatible(cls, target_branch, source_branch):
-        (scheme, user, password, host, port, path) = urlutils.parse_url(
-            target_branch.user_url)
-        return host in ('bazaar.launchpad.net', 'bazaar.staging.launchpad.net')
-
     def get_infotext(self):
         """Determine the initial comment for the merge proposal."""
         if self.commit_message is not None:

=== modified file 'propose.py'
--- propose.py	2018-09-09 14:31:18 +0000
+++ propose.py	2018-09-09 15:56:12 +0000
@@ -43,6 +43,16 @@
         self.url = url
 
 
+
+class UnsupportedHoster(errors.BzrError):
+
+    _fmt = "No supported hoster for %(branch)s."
+
+    def __init__(self, branch):
+        errors.BzrError.__init__(self)
+        self.branch = branch
+
+
 class ProposeMergeHooks(hooks.Hooks):
     """Hooks for proposing a merge on Launchpad."""
 
@@ -64,7 +74,7 @@
         self.url = url
 
 
-class MergeProposer(object):
+class MergeProposalBuilder(object):
     """Merge proposal creator.
 
     :param source_branch: Branch to propose for merging
@@ -107,7 +117,7 @@
     """A hosting site manager.
     """
 
-    def publish(self, base_branch, new_branch):
+    def publish(self, new_branch, base_branch):
         """Publish a branch to the site, derived from base_branch.
 
         :param base_branch: branch to derive the new branch from
@@ -121,7 +131,7 @@
 
         :param source_branch: Source branch
         :param target_branch: Target branch
-        :return: A MergeProposer object
+        :return: A MergeProposalBuilder object
         """
         raise NotImplementedError(self.get_proposer)
 
@@ -131,6 +141,14 @@
         raise NotImplementedError(cls.is_compatible)
 
 
+def get_hoster(branch):
+    """Find the hoster for a branch."""
+    for name, hoster_cls in hosters.items():
+        if hoster_cls.is_compatible(branch):
+            return hoster_cls()
+    raise UnsupportedHoster(branch)
+
+
 def get_proposer(branch, target_branch):
     """Create a merge proposal for branch to target_branch.
 
@@ -149,10 +167,21 @@
 proposers = registry.Registry()
 proposers.register_lazy(
         "launchpad", "breezy.plugins.propose.launchpad",
-        "LaunchpadMergeProposer")
-proposers.register_lazy(
-        "github", "breezy.plugins.propose.github",
-        "GitHubMergeProposer")
-proposers.register_lazy(
-        "gitlab", "breezy.plugins.propose.gitlab",
-        "GitlabMergeProposer")
+        "LaunchpadMergeProposalBuilder")
+proposers.register_lazy(
+        "github", "breezy.plugins.propose.github",
+        "GitHubMergeProposalBuilder")
+proposers.register_lazy(
+        "gitlab", "breezy.plugins.propose.gitlab",
+        "GitlabMergeProposalBuilder")
+
+hosters = registry.Registry()
+hosters.register_lazy(
+        "launchpad", "breezy.plugins.propose.launchpad",
+        "Launchpad")
+hosters.register_lazy(
+        "github", "breezy.plugins.propose.github",
+        "GitHub")
+hosters.register_lazy(
+        "gitlab", "breezy.plugins.propose.gitlab",
+        "GitLab")

